/*
 * Copyright (C) 2012 Michael Andersen <michael@steelcode.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.steelcode.pamflit;

import java.util.Scanner;

/**
 * Spawn a thread to wait and communicate with the device.
 * Although the current incarnation of PAMflit is a command line program
 * that could actually do without another thread, It is polite not to block
 * when called so that you can have a GUI or some other tasks running.
 * @author Michael Andersen <michael@steelcode.com>
 */
public class InteractiveConsole
{
    public InteractiveConsole()
    {
        //Can't think of anything we need to do here.
    }
    public void startBlocking()
    {
        NativeLink nl = NativeLink.getInstance();
        nl.connect(); //If this fails we will catch it below.
        byte [] buf = new byte[64];
        while(true)
        {
            int rv = nl.receiveFrame(buf);
            
            if (rv == 2) //Timeout: no problems, just nothing there for us
            {
                continue;
            }
            else if (rv != 0)
            {
                //If the device resets or is not in a part of the program where it 
                //is servicing the USB stack then we can momentarily disconnect.
                //Try automatically reconnect when we get a bad handle error
                try{
                    Thread.sleep(1000);
                }catch(Exception e){}
                if(nl.connect() == 0)
                {
                    System.err.println("Re-connected to device");
                }
                else
                {
                    System.err.println("Failed to reconnect");
                    
                    continue;
                }
            }
        //    System.err.print("+("+buf[0]+")");
            switch (buf[0])
            {
                case NativeLink.CODE_TX_INT:
                {
                    int v =  (int)buf[1] +
                            ((int)buf[2] << 8)  +
                            ((int)buf[3] << 16) +
                            ((int)buf[4] << 24);
                    String o = String.valueOf(v);
                    
                    for(int fieldwidth = buf[5];
                            fieldwidth > o.length();fieldwidth--)
                        System.err.print(" ");
                    
                    System.err.print(v);
                    
                    for(int fieldwidth = buf[5];
                            fieldwidth < -o.length();fieldwidth++)
                        System.err.print(" ");
                    
                    break;
                }
                case NativeLink.CODE_TX_CHAR:
                {
                    for(int fieldwidth = buf[3];
                            fieldwidth > 1;fieldwidth--)
                        System.err.print(" ");
                    
                    System.err.print((char)buf[1]);
                    
                    for(int fieldwidth = buf[3];
                            fieldwidth < -1;fieldwidth++)
                        System.err.print(" ");
                    break;
                }
                case NativeLink.CODE_TX_BOOL:
                {
                    String o = (buf[1]==0?"false":"true");
                    for(int fieldwidth = buf[3];
                            fieldwidth > o.length();fieldwidth--)
                        System.err.print(" ");
                    
                    System.err.print(o);
                    
                    for(int fieldwidth = buf[3];
                            fieldwidth < -o.length();fieldwidth++)
                        System.err.print(" ");
                    break;
                }
                case NativeLink.CODE_TX_STR:
                {
                    StringBuilder sb = new StringBuilder();
                    if(buf[1] > 32)
                    {   //The firmware packetizes strings into 32 byte chunks to overcome 
                        //the limitations of usb.
                        System.err.println("ERR: A somewhat unbelievable string came in :-(");
                        buf[1] = 32;
                    }
                    for (int i = 0; i < buf[1];i++)
                        sb.append((char)buf[i+2]);
                    System.err.print(sb.toString());
                    break;
                }   
                case NativeLink.CODE_RX_INT:
                {
                    System.err.print(" (int)>");
                    Scanner c = new Scanner(System.in);
                    boolean valid = false;
                    int v = 0;
                    while(!valid)
                    {
                        try
                        {
                            v = c.nextInt();
                            valid = true;
                        }
                        catch(Exception e)
                        {
                            System.err.println("Invalid integer!");
                        }
                    }
                    byte [] obuf = new byte [64];
                    obuf[0] = (byte)(v & 0xFF);
                    obuf[1] = (byte)((v >> 8)  & 0xFF);
                    obuf[2] = (byte)((v >> 16) & 0xFF);
                    obuf[3] = (byte)((v >> 24) & 0xFF);
                    nl.sendFrame(NativeLink.CODE_RX_INT, obuf);      
                    break;
                }
                case NativeLink.CODE_RX_BOOL:
                {
                    System.err.print(" (bool)>");
                    Scanner c = new Scanner(System.in);
                    boolean valid = false;
                    boolean v = false;
                    while(!valid)
                    {
                        try
                        {
                            v = c.nextBoolean();
                            valid = true;
                        }
                        catch(Exception e)
                        {
                            System.err.println("Invalid boolean!");
                        }
                    }
                    byte [] obuf = new byte [64];
                    obuf[0] = (byte)(v?1:0);
                    nl.sendFrame(NativeLink.CODE_RX_BOOL, obuf);      
                    break;
                }
                case NativeLink.CODE_RX_CHAR:
                {
                    System.err.print(" (char)>");
                    Scanner c = new Scanner(System.in);
                    boolean valid = false;
                    char v = ' ';
                    while(!valid)
                    {
                        try
                        {
                            v = c.nextLine().charAt(0);
                            valid = true;
                        }
                        catch(Exception e)
                        {
                            System.err.println("Invalid character!");
                        }
                    }
                    byte [] obuf = new byte [64];
                    obuf[0] = (byte)(v);
                    nl.sendFrame(NativeLink.CODE_RX_CHAR, obuf);      
                    break;
                }
                case NativeLink.CODE_DONE:
                {
                    //The ticks value is the system cycle clock (60Mhz) prescaled by 256.
                    int ticks =  (int)buf[2] +
                                ((int)buf[3] << 8)  +
                                ((int)buf[4] << 16) +
                                ((int)buf[5] << 24);
                    double ms = ticks* (256.0/60000.0);
                    System.err.println("\n=== TERMINATED ===");
                    System.err.println(" Program State: "+buf[1]); //These are the PS error codes. See the wiki
                    System.err.println(String.format(" Run time: %.3f ms",ms));
                    System.err.println("==================");
                    System.exit(0);
                    break;
                }
                case NativeLink.CODE_KLINE:
                { //Kill a line. We don't send it back to the board, but the board
                  //will block until it gets the 'ACK' token.
                    Scanner c = new Scanner(System.in);
                    System.err.print(" (\\n)>");
                    c.nextLine();
                    byte [] obuf = new byte [64];
                    nl.sendFrame(NativeLink.CODE_ACK, obuf); 
                    break;
                }
            }
        }
    }
    public void start()
    {
        Thread t = new Thread(new Runnable()
        {
            @Override
            public void run()
            {
                startBlocking();
            }
        });
        t.start();
    }
}
