/*
 * Copyright (C) 2012 Michael Andersen <michael@steelcode.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.steelcode.pamflit;

import java.io.File;

/**
 * Implements USB communication using JNI
 * Note that these methods can block for an appreciable amount of time
 * (on the order of 150ms) so you may want to not invoke them on the
 * EventQueue thread.
 * 
 * @author Michael Andersen <michael@steelcode.com>
 */
public class NativeLink 
{
    
    public static final int CODE_ERASE_BUFFER   = 0x25;
    public static final int CODE_FILL_BUFFER    = 0x23;
    public static final int CODE_FLASH_BUFFER   = 0x24;
    public static final int CODE_TX_INT         = 0x01|0x40;
    public static final int CODE_TX_CHAR        = 0x02|0x40;
    public static final int CODE_TX_BOOL        = 0x03|0x40;
    public static final int CODE_TX_STR         = 0x04|0x40;
    public static final int CODE_RX_INT         = 0x01;
    public static final int CODE_RX_CHAR        = 0x02;
    public static final int CODE_RX_BOOL        = 0x03;
    public static final int CODE_KLINE          = 0x05;
    public static final int CODE_ACK            = 0x06;
    public static final int CODE_DONE           = 0x10;

    private static NativeLink instance = null;
    public static NativeLink getInstance()
    {
        if (instance == null)
        {
            instance = new NativeLink();
        }
        return instance;
    }
    
    /**
     * Sends a payload (max 63 bytes) with the given type.
     * @param type What the payload is
     * @param payload The contents
     * @return 0 on success, nonzero on failure
     */
    public native int sendFrame(int type, byte [] payload);
    
    /**
     * Attempts to connect to a PAM
     * This function is safe to call if already connected
     * @return 0 if connected, 1 if unable to connect
     */
    public native int connect();
    
    /**
     * Attempts to receive a frame from a connected PAM.
     * @param buf A buffer of at least 63 bytes
     * @return less than 0 on error, the frame type (greater than zero) on success.
     */
    public native int receiveFrame(byte [] buf);
    
    /**
     * Some JNI incantations to load the native backing library.
     * The native library (in its various platform incarnations) is part of the
     * firmware repository but the binaries are copied to /native
     */
    private static boolean loadLib(String name, StringBuilder errors)
    {
        errors.append("\nAttempting native ").append(name).append(": \n");
        try
        {
            System.load(new File("native/"+name).getAbsolutePath());
            errors.append("success\n");
            return true;
        }
        catch(UnsatisfiedLinkError e)
        {errors.append(e.getMessage()).append("\n");}
        try
        {
            System.load(new File("PAMflit/native/"+name).getAbsolutePath());
            errors.append("success\n");
            return true;
        }
        catch(UnsatisfiedLinkError e)
        {errors.append(e.getMessage()).append("\n");}
        return false;
    }
    public static void initNatives() 
    {
        StringBuilder errors = new StringBuilder();
        if(loadLib("libpam32.dll",errors))
        {
            errors.append("Loaded 32 bit windows natives\n");
        }
        else if(loadLib("libpam64.dll",errors))
        {
            errors.append("Loaded 64 bit windows natives\n");
        }
        else if(loadLib("libpam64.so",errors))
        {
            errors.append("Loading 64 bit linux natives\n");
            if(!loadLib("libusb64-1.0.so",errors))
            {
                errors.append("Failed to load libusb64!\n");
                System.err.println("Aborting, errors follow:");
                System.err.println(errors.toString());
                System.exit(4);
            }
        }
        else if (loadLib("libpam32.so",errors))
        {
            errors.append("Loading 32 bit linux natives\n");
            if(!loadLib("libusb32-1.0.so",errors))
            {
                errors.append("Failed to load libusb32!\n");
                System.err.println("Aborting, errors follow:");
                System.err.println(errors.toString());
                System.exit(4);
            } 
        }else
        {
            errors.append("No native libraries loaded :/\n");
            System.err.println("Aborting, errors follow:");
            System.err.println(errors.toString());
            System.exit(5);
        }
        System.err.println(".:: A native interface loaded correctly ::.");
    }
}
