/*
 * Copyright (C) 2012 Michael Andersen <michael@steelcode.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.steelcode.pamflit;

import java.util.ArrayList;
import java.util.Arrays;

/**
 * A program that can be written to a PAM module
 * 
 * @author Michael Andersen <michael@steelcode.com>
 */
public class ProgramImage 
{
    private int [] memory;
    /**
     * Construct a program image of the default size
     */
    public ProgramImage()
    {
        this(8192);
    }
    
    /**
     * Construct a program image of a specific size
     */
    public ProgramImage(int size)
    {
        this.memory = new int [size];
    }

    /**
     * Write a set of instructions to the image.
     * @param org Where to start from (often 0)
     * @param codes The codes to write
     */
    public void writeCodes(int org, ArrayList<Integer> codes)
    {
        for(Integer i : codes)
        {
            memory[org] = i;
            org++;
        }
    }
    
    /**
     * Write a set of instructions to the image.
     * @param org Where to start from (often 0)
     * @param codes The codes to write
     */
    public void writeCodes(int org, int [] codes)
    {
        for(Integer i : codes)
        {
            memory[org] = i;
            org++;
        }
    }
    
    /**
     * Program the current image to the device
     * @return 0 on success nonzero on failure
     */
    public int programToDevice()
    {
        /*
         * Writing to the nonvolatile program memory consists of three phases:
         *  1) Clear existing program memory buffer
         *  2) Upload the image 16 instructions at a time
         *  3) Trigger the internal flash erase/reprogram cycle
         * 
         * This is because internally, PAM is like an SSD or a flash drive
         * where you can only write in pages (of 1024 instructions). The device
         * repurposes the RAM as a buffer in step 1, fills it in step 2 and then
         * writes it in pages in step 3.
         */
        NativeLink nl = NativeLink.getInstance();
        if (nl.connect() != 0)
            return -1;     
        byte [] buf = new byte[64];

        try
        {
        Thread.sleep(1000);
        }catch(Exception e){}
        //Step 1
        //buf is unused but required
        if(nl.sendFrame(NativeLink.CODE_ERASE_BUFFER, buf) != 0)
            return -2;
        
        //Step 2
        for (int i = 0;i<memory.length/16;i++)
        {
          //  System.err.println("Programming code segment "+i);
            //PAM expects the first two bytes to be a little endian uint16_t
            //of where to put this payload in the RAM buffer
            buf[0] = (byte)(i & 0xFF);
            buf[1] = (byte)(i >> 8);
            for(int k = 0; k<16;k++)
            {
                //Each element is also packed little endian
                buf[2+k*2] = (byte)(memory[i*16+k] & 0xFF);
                buf[3+k*2] = (byte)(memory[i*16+k] >> 8);
            }
            
            int rv = nl.sendFrame(NativeLink.CODE_FILL_BUFFER, buf);
            if (rv != 0)
            {
                System.out.println("Failed to program code segment "+i);
                return -3;
            }
        }
        
        //Step 3
        //buf is unused again
        if(nl.sendFrame(NativeLink.CODE_FLASH_BUFFER, buf) != 0)
        {
            System.out.println("Failed to initialize flash sequence");
            return -4;
        }
        
        //Success
        return 0;
    }
}
